@@ -15,7 +15,7 @@
 -- and extracting files from zip archives.
 --
 -- Certain simplifying assumptions are made about the zip archives: in
--- particular, there is no support for encryption, zip files that span
+-- particular, there is no support for strong encryption, zip files that span
 -- multiple disks, ZIP64, OS-specific file attributes, or compression
 -- methods other than Deflate.  However, the library should be able to
 -- read the most common zip archives, and the archives it produces should
@@ -35,6 +35,7 @@ module Codec.Archive.Zip
          Archive (..)
        , Entry (..)
        , CompressionMethod (..)
+       , EncryptionMethod (..)
        , ZipOption (..)
        , ZipException (..)
        , emptyArchive
@@ -48,6 +49,8 @@ module Codec.Archive.Zip
        , deleteEntryFromArchive
        , findEntryByPath
        , fromEntry
+       , fromEncryptedEntry
+       , isEncryptedEntry
        , toEntry
 #ifndef _WINDOWS
        , isEntrySymbolicLink
@@ -70,20 +73,20 @@ import Data.Time.Calendar ( toGregorian, fromGregorian )
 import Data.Time.Clock ( UTCTime(..) )
 import Data.Time.Clock.POSIX ( posixSecondsToUTCTime, utcTimeToPOSIXSeconds )
 import Data.Time.LocalTime ( TimeOfDay(..), timeToTimeOfDay )
-import Data.Bits ( shiftL, shiftR, (.&.) )
+import Data.Bits ( shiftL, shiftR, (.&.), (.|.), xor, testBit )
 import Data.Binary
 import Data.Binary.Get
 import Data.Binary.Put
-import Data.List ( nub, find, intercalate, partition)
+import Data.List (nub, find, intercalate)
 import Data.Data (Data)
-import Data.Maybe (fromJust)
 import Data.Typeable (Typeable)
 import Text.Printf
 import System.FilePath
-import System.Directory ( doesDirectoryExist, getDirectoryContents, createDirectoryIfMissing )
-import Control.Monad ( when, unless, zipWithM )
+import System.Directory
+       (doesDirectoryExist, getDirectoryContents,
+        createDirectoryIfMissing, getModificationTime)
+import Control.Monad ( when, unless, zipWithM_ )
 import qualified Control.Exception as E
-import System.Directory ( getModificationTime )
 import System.IO ( stderr, hPutStrLn )
 import qualified Data.Digest.CRC32 as CRC32
 import qualified Data.Map as M
@@ -93,12 +96,14 @@ import Control.Applicative
 #ifndef _WINDOWS
 import System.Posix.Files ( setFileTimes, setFileMode, fileMode, getSymbolicLinkStatus, symbolicLinkMode, readSymbolicLink, isSymbolicLink, unionFileModes, createSymbolicLink )
 import System.Posix.Types ( CMode(..) )
+import Data.List (partition)
+import Data.Maybe (fromJust)
 #endif
 
 -- from bytestring
 import qualified Data.ByteString as S
 import qualified Data.ByteString.Lazy as B
-import qualified Data.ByteString.Lazy.Char8 as C (pack, unpack)
+import qualified Data.ByteString.Lazy.Char8 as C
 
 -- text
 import qualified Data.Text.Lazy as TL
@@ -107,13 +112,6 @@ import qualified Data.Text.Lazy.Encoding as TL
 -- from zlib
 import qualified Codec.Compression.Zlib.Raw as Zlib
 
-versionMadeBy :: Word16
-#ifdef _WINDOWS
-versionMadeBy = 0x0000 -- FAT/VFAT/VFAT32 file attributes
-#else
-versionMadeBy = 0x0300 -- UNIX file attributes
-#endif
-
 #if !MIN_VERSION_binary(0, 6, 0)
 manySig :: Word32 -> Get a -> Get [a]
 manySig sig p = do
@@ -145,6 +143,7 @@ instance Binary Archive where
 data Entry = Entry
                { eRelativePath            :: FilePath            -- ^ Relative path, using '/' as separator
                , eCompressionMethod       :: CompressionMethod   -- ^ Compression method
+               , eEncryptionMethod        :: EncryptionMethod    -- ^ Encryption method
                , eLastModified            :: Integer             -- ^ Modification time (seconds since unix epoch)
                , eCRC32                   :: Word32              -- ^ CRC32 checksum
                , eCompressedSize          :: Word32              -- ^ Compressed size in bytes
@@ -162,6 +161,15 @@ data CompressionMethod = Deflate
                        | NoCompression
                        deriving (Read, Show, Eq)
 
+data EncryptionMethod = NoEncryption            -- ^ Entry is not encrypted
+                      | PKWAREEncryption Word8  -- ^ Entry is encrypted with the traditional PKWARE encryption
+                      deriving (Read, Show, Eq)
+
+-- | The way the password should be verified during entry decryption
+data PKWAREVerificationType = CheckTimeByte
+                            | CheckCRCByte
+                            deriving (Read, Show, Eq)
+
 -- | Options for 'addFilesToArchive' and 'extractFilesFromArchive'.
 data ZipOption = OptRecursive               -- ^ Recurse into directories when adding files
                | OptVerbose                 -- ^ Print information to stderr
@@ -171,8 +179,10 @@ data ZipOption = OptRecursive               -- ^ Recurse into directories when a
                deriving (Read, Show, Eq)
 
 data ZipException =
-  CRC32Mismatch FilePath
-  deriving (Show, Typeable, Data)
+    CRC32Mismatch FilePath
+  | UnsafePath FilePath
+  | CannotWriteEncryptedEntry FilePath
+  deriving (Show, Typeable, Data, Eq)
 
 instance E.Exception ZipException
 
@@ -207,7 +217,7 @@ fromArchive = encode
 
 -- | Returns a list of files in a zip archive.
 filesInArchive :: Archive -> [FilePath]
-filesInArchive = (map eRelativePath) . zEntries
+filesInArchive = map eRelativePath . zEntries
 
 -- | Adds an entry to a zip archive, or updates an existing entry.
 addEntryToArchive :: Entry -> Archive -> Archive
@@ -232,6 +242,18 @@ fromEntry :: Entry -> B.ByteString
 fromEntry entry =
   decompressData (eCompressionMethod entry) (eCompressedData entry)
 
+-- | Returns decrypted and uncompressed contents of zip entry.
+fromEncryptedEntry :: String -> Entry -> Maybe B.ByteString
+fromEncryptedEntry password entry =
+  decompressData (eCompressionMethod entry) <$> decryptData password (eEncryptionMethod entry) (eCompressedData entry)
+
+-- | Check if an 'Entry' is encrypted
+isEncryptedEntry :: Entry -> Bool
+isEncryptedEntry entry =
+  case eEncryptionMethod entry of
+    (PKWAREEncryption _) -> True
+    _ -> False
+
 -- | Create an 'Entry' with specified file path, modification time, and contents.
 toEntry :: FilePath         -- ^ File path for entry
         -> Integer          -- ^ Modification time for entry (seconds since unix epoch)
@@ -249,6 +271,7 @@ toEntry path modtime contents =
       crc32 = CRC32.crc32 contents
   in  Entry { eRelativePath            = normalizePath path
             , eCompressionMethod       = compressionMethod
+            , eEncryptionMethod        = NoEncryption
             , eLastModified            = modtime
             , eCRC32                   = crc32
             , eCompressedSize          = fromIntegral finalSize
@@ -293,13 +316,12 @@ readEntry opts path = do
                         return B.empty
                       else
                         B.fromStrict <$> S.readFile path
-  modEpochTime <- fmap (floor . utcTimeToPOSIXSeconds)
-                   $ getModificationTime path
+  modEpochTime <- (floor . utcTimeToPOSIXSeconds) <$> getModificationTime path
   let entry = toEntry path' modEpochTime contents
 
   entryE <-
 #ifdef _WINDOWS
-        return $ entry
+        return $ entry { eVersionMadeBy = 0x0000 } -- FAT/VFAT/VFAT32 file attributes
 #else
         do
            let fm = if isSymLink
@@ -308,7 +330,7 @@ readEntry opts path = do
 
            let modes = fromIntegral $ shiftL (toInteger fm) 16
            return $ entry { eExternalFileAttributes = modes,
-                            eVersionMadeBy = versionMadeBy }
+                            eVersionMadeBy = 0x0300 } -- UNIX file attributes
 #endif
 
   when (OptVerbose `elem` opts) $ do
@@ -325,9 +347,18 @@ readEntry opts path = do
 -- does not match the uncompressed data.
 writeEntry :: [ZipOption] -> Entry -> IO ()
 writeEntry opts entry = do
-  let path = case [d | OptDestination d <- opts] of
-                  (x:_) -> x </> eRelativePath entry
-                  _     -> eRelativePath entry
+  when (isEncryptedEntry entry) $
+    E.throwIO $ CannotWriteEncryptedEntry (eRelativePath entry)
+  let relpath = eRelativePath entry
+  let isUnsafePath = ".." `elem` splitDirectories relpath
+  when isUnsafePath $
+    E.throwIO $ UnsafePath relpath
+  path <- case [d | OptDestination d <- opts] of
+             (x:_) -> return (x </> relpath)
+             _ | isAbsolute relpath
+                   -> E.throwIO $ UnsafePath relpath
+               | otherwise
+                   -> return relpath
   -- create directories if needed
   let dir = takeDirectory path
   exists <- doesDirectoryExist dir
@@ -335,10 +366,10 @@ writeEntry opts entry = do
     createDirectoryIfMissing True dir
     when (OptVerbose `elem` opts) $
       hPutStrLn stderr $ "  creating: " ++ dir
-  if length path > 0 && last path == '/' -- path is a directory
+  if not (null path) && last path == '/' -- path is a directory
      then return ()
      else do
-       when (OptVerbose `elem` opts) $ do
+       when (OptVerbose `elem` opts) $
          hPutStrLn stderr $ case eCompressionMethod entry of
                                  Deflate       -> " inflating: " ++ path
                                  NoCompression -> "extracting: " ++ path
@@ -361,11 +392,11 @@ writeEntry opts entry = do
 -- the options do not contain 'OptPreserveSymbolicLinks`, this
 -- function behaves like `writeEntry`.
 writeSymbolicLinkEntry :: [ZipOption] -> Entry -> IO ()
-writeSymbolicLinkEntry opts entry = do
+writeSymbolicLinkEntry opts entry =
   if OptPreserveSymbolicLinks `notElem` opts
      then writeEntry opts entry
      else do
-        if (isEntrySymbolicLink entry)
+        if isEntrySymbolicLink entry
            then do
              let prefixPath = case [d | OptDestination d <- opts] of
                                    (x:_) -> x
@@ -403,7 +434,7 @@ addFilesToArchive opts archive files = do
 #ifdef _WINDOWS
                          then mapM getDirectoryContentsRecursive files >>= return . nub . concat
 #else
-                         then mapM (getDirectoryContentsRecursive' opts) files >>= return . nub . concat
+                         then nub . concat <$> mapM (getDirectoryContentsRecursive' opts) files
 #endif
                          else return files
   entries <- mapM (readEntry opts) filesAndChildren
@@ -413,18 +444,20 @@ addFilesToArchive opts archive files = do
 -- as needed.  If 'OptVerbose' is specified, print messages to stderr.
 -- Note that the last-modified time is set correctly only in POSIX,
 -- not in Windows.
+-- This function fails if encrypted entries are present
 extractFilesFromArchive :: [ZipOption] -> Archive -> IO ()
 extractFilesFromArchive opts archive = do
+  let entries = zEntries archive
   if OptPreserveSymbolicLinks `elem` opts
     then do
 #ifdef _WINDOWS
-      mapM_ (writeEntry opts) $ zEntries archive
+      mapM_ (writeEntry opts) entries
 #else
-      let (symbolicLinkEntries, nonSymbolicLinkEntries) = partition isEntrySymbolicLink $ zEntries archive
-      mapM_ (writeEntry opts) $ nonSymbolicLinkEntries
-      mapM_ (writeSymbolicLinkEntry opts) $ symbolicLinkEntries
+      let (symbolicLinkEntries, nonSymbolicLinkEntries) = partition isEntrySymbolicLink entries
+      mapM_ (writeEntry opts) nonSymbolicLinkEntries
+      mapM_ (writeSymbolicLinkEntry opts) symbolicLinkEntries
 #endif
-    else mapM_ (writeEntry opts) $ zEntries archive
+    else mapM_ (writeEntry opts) entries
 
 --------------------------------------------------------------------------------
 -- Internal functions for reading and writing zip binary format.
@@ -453,6 +486,39 @@ decompressData :: CompressionMethod -> B.ByteString -> B.ByteString
 decompressData Deflate       = Zlib.decompress
 decompressData NoCompression = id
 
+-- | Decrypt a lazy bytestring
+-- Returns Nothing if password is incorrect
+decryptData :: String -> EncryptionMethod -> B.ByteString -> Maybe B.ByteString
+decryptData _ NoEncryption s = Just s
+decryptData password (PKWAREEncryption controlByte) s =
+  let headerlen = 12
+      initKeys = (305419896, 591751049, 878082192)
+      startKeys = B.foldl pkwareUpdateKeys initKeys (C.pack password)
+      (header, content) = B.splitAt headerlen $ snd $ B.mapAccumL pkwareDecryptByte startKeys s
+  in if B.last header == controlByte
+        then Just content
+        else Nothing
+
+-- | PKWARE decryption context
+type DecryptionCtx = (Word32, Word32, Word32)
+
+-- | An interation of the PKWARE decryption algorithm
+pkwareDecryptByte :: DecryptionCtx -> Word8 -> (DecryptionCtx, Word8)
+pkwareDecryptByte keys@(_, _, key2) inB =
+  let tmp = key2 .|. 2
+      tmp' = fromIntegral ((tmp * (tmp `xor` 1)) `shiftR` 8) :: Word8
+      outB = inB `xor` tmp'
+  in (pkwareUpdateKeys keys outB, outB)
+
+-- | Update decryption keys after a decrypted byte
+pkwareUpdateKeys :: DecryptionCtx -> Word8 -> DecryptionCtx
+pkwareUpdateKeys (key0, key1, key2) inB =
+  let key0' = CRC32.crc32Update (key0 `xor` 0xffffffff) [inB] `xor` 0xffffffff
+      key1' = (key1 + (key0' .&. 0xff)) * 134775813 + 1
+      key1Byte = fromIntegral (key1' `shiftR` 24) :: Word8
+      key2' = CRC32.crc32Update (key2 `xor` 0xffffffff) [key1Byte] `xor` 0xffffffff
+  in (key0', key1', key2')
+
 -- | Calculate compression ratio for an entry (for verbose output).
 compressionRatio :: Entry -> Float
 compressionRatio entry =
@@ -482,10 +548,10 @@ epochTimeToMSDOSDateTime epochtime | epochtime < minMSDOSDateTime =
   epochTimeToMSDOSDateTime minMSDOSDateTime
   -- if time is earlier than minimum DOS datetime, return minimum
 epochTimeToMSDOSDateTime epochtime =
-  let 
+  let
     UTCTime
       (toGregorian -> (fromInteger -> year, month, day))
-      (timeToTimeOfDay -> (TimeOfDay hour minutes (floor -> sec))) 
+      (timeToTimeOfDay -> (TimeOfDay hour minutes (floor -> sec)))
       = posixSecondsToUTCTime (fromIntegral epochtime)
 
     dosTime = toEnum $ (sec `div` 2) + shiftL minutes 5 + shiftL hour 11
@@ -494,19 +560,19 @@ epochTimeToMSDOSDateTime epochtime =
 
 -- | Convert a MSDOS datetime to a 'ClockTime'.
 msDOSDateTimeToEpochTime :: MSDOSDateTime -> Integer
-msDOSDateTimeToEpochTime (MSDOSDateTime {msDOSDate = dosDate, msDOSTime = dosTime}) =
+msDOSDateTimeToEpochTime MSDOSDateTime {msDOSDate = dosDate, msDOSTime = dosTime} =
   let seconds = fromIntegral $ 2 * (dosTime .&. 0O37)
-      minutes = fromIntegral $ (shiftR dosTime 5) .&. 0O77
+      minutes = fromIntegral $ shiftR dosTime 5 .&. 0O77
       hour    = fromIntegral $ shiftR dosTime 11
       day     = fromIntegral $ dosDate .&. 0O37
-      month   = fromIntegral $ ((shiftR dosDate 5) .&. 0O17)
+      month   = fromIntegral ((shiftR dosDate 5) .&. 0O17)
       year    = fromIntegral $ shiftR dosDate 9
       utc = UTCTime (fromGregorian (1980 + year) month day) (3600 * hour + 60 * minutes + seconds)
   in floor (utcTimeToPOSIXSeconds utc)
 
 #ifndef _WINDOWS
 getDirectoryContentsRecursive' :: [ZipOption] -> FilePath -> IO [FilePath]
-getDirectoryContentsRecursive' opts path = do
+getDirectoryContentsRecursive' opts path =
   if OptPreserveSymbolicLinks `elem` opts
      then do
        isDir <- doesDirectoryExist path
@@ -538,10 +604,10 @@ getDirectoryContentsRecursivelyBy exploreMethod path = do
 
 
 setFileTimeStamp :: FilePath -> Integer -> IO ()
-setFileTimeStamp file epochtime = do
 #ifdef _WINDOWS
-  return ()  -- TODO - figure out how to set the timestamp on Windows
+setFileTimeStamp _ _ = return () -- TODO: figure out how to set the timestamp on Windows
 #else
+setFileTimeStamp file epochtime = do
   let epochtime' = fromInteger epochtime
   setFileTimes file epochtime' epochtime'
 #endif
@@ -615,7 +681,7 @@ getArchive = do
   skip 4 -- offset of central directory
   commentLength <- getWord16le
   zipComment <- getLazyByteString (toEnum $ fromEnum commentLength)
-  return $ Archive
+  return Archive
            { zEntries                = files
            , zSignature              = digSig
            , zComment                = zipComment
@@ -627,7 +693,7 @@ putArchive archive = do
   let localFileSizes = map localFileSize $ zEntries archive
   let offsets = scanl (+) 0 localFileSizes
   let cdOffset = last offsets
-  _ <- zipWithM putFileHeader offsets (zEntries archive)
+  _ <- zipWithM_ putFileHeader offsets (zEntries archive)
   putDigitalSignature $ zSignature archive
   putWord32le 0x06054b50
   putWord16le 0 -- disk number
@@ -705,7 +771,7 @@ getLocalFile = do
               cs <- getWord32le  -- compressed size
               skip 4 -- uncompressed size
               if fromIntegral cs == B.length raw
-                 then return $ raw
+                 then return raw
                  else fail "Content size mismatch in data descriptor record"
   return (fromIntegral offset, compressedData)
 
@@ -824,7 +890,7 @@ getFileHeader locals = do
   skip 1 -- upper byte indicates OS part of "version needed to extract"
   unless (versionNeededToExtract <= 20) $
     fail "This archive requires zip >= 2.0 to extract."
-  skip 2 -- general purpose bit flag
+  bitflag <- getWord16le
   rawCompressionMethod <- getWord16le
   compressionMethod <- case rawCompressionMethod of
                         0 -> return NoCompression
@@ -833,6 +899,12 @@ getFileHeader locals = do
   lastModFileTime <- getWord16le
   lastModFileDate <- getWord16le
   crc32 <- getWord32le
+  encryptionMethod <- case (testBit bitflag 0, testBit bitflag 3, testBit bitflag 6) of
+                        (False, _, _) -> return NoEncryption
+                        (True, False, False) -> return $ PKWAREEncryption (fromIntegral (crc32 `shiftR` 24))
+                        (True, True, False) -> return $ PKWAREEncryption (fromIntegral (lastModFileTime `shiftR` 8))
+                        (True, _, True) -> fail "Strong encryption is not supported"
+
   compressedSize <- getWord32le
   uncompressedSize <- getWord32le
   fileNameLength <- getWord16le
@@ -845,13 +917,14 @@ getFileHeader locals = do
   fileName <- getLazyByteString (toEnum $ fromEnum fileNameLength)
   extraField <- getLazyByteString (toEnum $ fromEnum extraFieldLength)
   fileComment <- getLazyByteString (toEnum $ fromEnum fileCommentLength)
-  compressedData <- case (M.lookup relativeOffset locals) of
+  compressedData <- case M.lookup relativeOffset locals of
                     Just x  -> return x
                     Nothing -> fail $ "Unable to find data at offset " ++
                                         show relativeOffset
-  return $ Entry
+  return Entry
             { eRelativePath            = toString fileName
             , eCompressionMethod       = compressionMethod
+            , eEncryptionMethod        = encryptionMethod
             , eLastModified            = msDOSDateTimeToEpochTime $
                                          MSDOSDateTime { msDOSDate = lastModFileDate,
                                                          msDOSTime = lastModFileTime }
