@@ -5,11 +5,14 @@
 
 import Codec.Archive.Zip
 import Control.Applicative
+import Control.Monad (unless)
+import Control.Exception (try)
 import System.Directory hiding (isSymbolicLink)
 import Test.HUnit.Base
 import Test.HUnit.Text
 import qualified Data.ByteString.Char8 as BS
 import qualified Data.ByteString.Lazy as BL
+import qualified Data.ByteString.Lazy.Char8 as BLC
 import System.Exit
 import System.IO.Temp (withTempDirectory)
 
@@ -48,7 +51,13 @@ createTestDirectoryWithSymlinks prefixDir  baseDir = do
 
 main :: IO Counts
 main = withTempDirectory "." "test-zip-archive." $ \tmpDir -> do
-  res   <- runTestTT $ TestList $ map (\f -> f tmpDir)
+#ifndef _WINDOWS
+  ec <- rawSystem "which" ["unzip"]
+  let unzipInPath = ec == ExitSuccess
+  unless unzipInPath $
+    putStrLn "\n\nunzip is not in path; skipping testArchiveAndUnzip\n"
+#endif
+  res   <- runTestTT $ TestList $ map (\f -> f tmpDir) $
                                 [ testReadWriteArchive
                                 , testReadExternalZip
                                 , testFromToArchive
@@ -56,13 +65,19 @@ main = withTempDirectory "." "test-zip-archive." $ \tmpDir -> do
                                 , testAddFilesOptions
                                 , testDeleteEntries
                                 , testExtractFiles
+                                , testExtractFilesFailOnEncrypted
+                                , testPasswordProtectedRead
+                                , testIncorrectPasswordRead
+                                , testEvilPath
 #ifndef _WINDOWS
                                 , testExtractFilesWithPosixAttrs
                                 , testArchiveExtractSymlinks
                                 , testExtractExternalZipWithSymlinks
-                                , testArchiveAndUnzip
 #endif
                                 ]
+#ifndef _WINDOWS
+                                ++ [testArchiveAndUnzip | unzipInPath]
+#endif
   exitWith $ case (failures res + errors res) of
                      0 -> ExitSuccess
                      n -> ExitFailure n
@@ -85,7 +100,10 @@ testReadExternalZip _tmpDir = TestCase $ do
   bContents <- BL.readFile "tests/test4/b.bin"
   case findEntryByPath "test4/b.bin" archive of
        Nothing  -> assertFailure "test4/b.bin not found in archive"
-       Just f   -> assertEqual "for contents of test4/b.bin in archive"
+       Just f   -> do
+                    assertEqual "for text4/b.bin file entry"
+                      NoEncryption (eEncryptionMethod f)
+                    assertEqual "for contents of test4/b.bin in archive"
                       bContents (fromEntry f)
   case findEntryByPath "test4/" archive of
        Nothing  -> assertFailure "test4/ not found in archive"
@@ -136,6 +154,14 @@ testDeleteEntries _tmpDir = TestCase $ do
   let archive3 = deleteEntryFromArchive "src" archive2
   assertEqual "for deleteFilesFromArchive" emptyArchive archive3
 
+testEvilPath :: FilePath -> Test
+testEvilPath _tmpDir = TestCase $ do
+  archive <- toArchive <$> BL.readFile "tests/zip_with_evil_path.zip"
+  result <- try $ extractFilesFromArchive [] archive :: IO (Either ZipException ())
+  case result of
+    Left err -> assertBool "Wrong exception" $ err == UnsafePath "../evil"
+    Right _ -> assertFailure "extractFilesFromArchive should have failed"
+
 testExtractFiles :: FilePath -> Test
 testExtractFiles tmpDir = TestCase $ do
   createDirectory (tmpDir </> "dir1")
@@ -152,6 +178,41 @@ testExtractFiles tmpDir = TestCase $ do
   assertEqual ("contents of " </> tmpDir </> "dir1/hi") hiMsg hi
   assertEqual ("contents of " </> tmpDir </> "dir1/dir2/hello") helloMsg hello
 
+testExtractFilesFailOnEncrypted :: FilePath -> Test
+testExtractFilesFailOnEncrypted tmpDir = TestCase $ do
+  let dir = tmpDir </> "fail-encrypted"
+  createDirectory dir
+
+  archive <- toArchive <$> BL.readFile "tests/zip_with_password.zip"
+  result <- try $ extractFilesFromArchive [OptDestination dir] archive :: IO (Either ZipException ())
+  removeDirectoryRecursive dir
+
+  case result of
+    Left err -> assertBool "Wrong exception" $ err == CannotWriteEncryptedEntry "test.txt"
+    Right _ -> assertFailure "extractFilesFromArchive should have failed"
+
+testPasswordProtectedRead :: FilePath -> Test
+testPasswordProtectedRead _tmpDir = TestCase $ do
+  archive <- toArchive <$> BL.readFile "tests/zip_with_password.zip"
+
+  assertEqual "for results of filesInArchive" ["test.txt"] (filesInArchive archive)
+  case findEntryByPath "test.txt" archive of
+       Nothing  -> assertFailure "test.txt not found in archive"
+       Just f   -> do
+            assertBool "for encrypted test.txt file entry"
+              (isEncryptedEntry f)
+            assertEqual "for contents of test.txt in archive"
+              (Just $ BLC.pack "SUCCESS\n") (fromEncryptedEntry "s3cr3t" f)
+
+testIncorrectPasswordRead :: FilePath -> Test
+testIncorrectPasswordRead _tmpDir = TestCase $ do
+  archive <- toArchive <$> BL.readFile "tests/zip_with_password.zip"
+  case findEntryByPath "test.txt" archive of
+       Nothing  -> assertFailure "test.txt not found in archive"
+       Just f   -> do
+            assertEqual "for contents of test.txt in archive"
+              Nothing (fromEncryptedEntry "INCORRECT" f)
+
 #ifndef _WINDOWS
 
 testExtractFilesWithPosixAttrs :: FilePath -> Test
